
module adapter_instr #(
    parameter AXI_ADDR_WIDTH = 0,
    parameter COUNTER_WIDTH = 0,
    parameter FIFO_LEN = 0,
    parameter MAX_EVENT_BUF_LEN = 0, //Must be power of two
    parameter WAIT_FOR_AVAIL_SLOT = 0
) (
    input clk,
    input rstn,
    //Event in stream
    input event_in_tvalid,
    output event_in_tready,
    input [111:0] event_in_tdata,
    //Instr buffer AXI
    //AR
    output instr_buf_arvalid,
    input instr_buf_arready,
    output [AXI_ADDR_WIDTH-1:0] instr_buf_araddr,
    output [7:0] instr_buf_arlen,
    output [2:0] instr_buf_arsize,
    output [1:0] instr_buf_arburst,
    output [3:0] instr_buf_arcache,
    //R
    input instr_buf_rvalid,
    output instr_buf_rready,
    input [63:0] instr_buf_rdata,
    input instr_buf_rlast,
    input [1:0] instr_buf_rresp,
    //AW
    output instr_buf_awvalid,
    input instr_buf_awready,
    output [AXI_ADDR_WIDTH-1:0] instr_buf_awaddr,
    output [7:0] instr_buf_awlen,
    output [2:0] instr_buf_awsize,
    output [1:0] instr_buf_awburst,
    output [3:0] instr_buf_awcache,
    //W
    output instr_buf_wvalid,
    input instr_buf_wready,
    output logic [63:0] instr_buf_wdata,
    output [7:0] instr_buf_wstrb,
    output instr_buf_wlast,
    //B
    input instr_buf_bvalid,
    output instr_buf_bready,
    input [1:0] instr_buf_bresp
);

    localparam EVENT_TYPE_INVALID = 2'd3;
    localparam EVENT_INSTEVLOST = 32'd82;
    localparam EVENT_TYPE_POINT = 8'd2;

    localparam EVENT_BUF_BITS = $clog2(MAX_EVENT_BUF_LEN);
    localparam EVENT_BUF_ONE = {{EVENT_BUF_BITS{1'b0}}, 1'b1};
    localparam EVENT_BUF_ADDR_ONE = {{EVENT_BUF_BITS-1{1'b0}}, 1'b1};

    typedef enum bit [2:0] {
        IDLE,
        READ_BUFFER_AR,
        WAIT_READ_BUFFER,
        CHECK_AVAIL_SLOTS,
        WRITE_BUFFER,
        UPDATE_SLOTS,
        WAIT_WRESP
    } State_t;

    State_t state;
    wire fifo_tvalid;
    wire fifo_tready;
    wire [104:0] fifo_tdata;
    wire [COUNTER_WIDTH-1:0] fifo_timestamp;
    reg [COUNTER_WIDTH-1:0] timestamp;
    reg [EVENT_BUF_BITS:0] slots;
    reg [EVENT_BUF_BITS-1:0] current_slot;
    reg [EVENT_BUF_BITS-1:0] read_slot;
    reg [EVENT_BUF_BITS:0] avail_slots;
    reg [1:0] write_buffer_count;
    reg [AXI_ADDR_WIDTH-3-1:0] instr_buffer_addr;
    wire [AXI_ADDR_WIDTH-3-1:0] instr_buffer_addr_read;
    wire [AXI_ADDR_WIDTH-3-1:0] instr_buffer_addr_write;
    reg [63:0] instr_event_data;
    reg [39:0] type_and_id;
    reg overflow;
    reg data_sent;
    reg addr_sent;
    reg [1:0] bcount;

    assign instr_buffer_addr_read = instr_buffer_addr + read_slot*3'd3 + 2'd2;
    assign instr_buf_araddr = {instr_buffer_addr_read, 3'd0};
    assign instr_buf_arvalid = state == READ_BUFFER_AR;
    assign instr_buf_arlen = 8'd0;
    assign instr_buf_arsize = 3'd3;
    assign instr_buf_arcache = 4'b0010; //transaction modifiable
    assign instr_buf_arburst = 2'b01; //INCR
    assign instr_buf_rready = state == WAIT_READ_BUFFER;

    assign instr_buffer_addr_write = instr_buffer_addr + current_slot*3'd3 + write_buffer_count;
    assign instr_buf_awaddr = {instr_buffer_addr_write, 3'd0};
    assign instr_buf_awvalid = state == WRITE_BUFFER && !addr_sent;
    assign instr_buf_awlen = 8'd0;
    assign instr_buf_awsize = 3'd3;
    assign instr_buf_awcache = 4'b0010;
    assign instr_buf_awburst = 2'b01;
    assign instr_buf_wvalid = state == WRITE_BUFFER && !data_sent;
    assign instr_buf_wstrb = 8'hFF;
    assign instr_buf_wlast = 1'b1;

    assign instr_buf_bready = 1'b1;

    assign fifo_tready = state == IDLE;

    adapter_instr_fifo #(
        .COUNTER_WIDTH(COUNTER_WIDTH),
        .LEN(FIFO_LEN)
    ) fifo_I (
        .clk(clk),
        .rstn(rstn),
        .si_tvalid(event_in_tvalid),
        .si_tready(event_in_tready),
        .si_tdata(event_in_tdata[104:0]),
        .so_tvalid(fifo_tvalid),
        .so_tready(fifo_tready),
        .so_tdata(fifo_tdata),
        .so_timestamp(fifo_timestamp)
    );

    always_comb begin
        instr_buf_wdata = 64'd0;

        case (write_buffer_count)

            2'd2: begin
                if (overflow) begin
                    instr_buf_wdata[39:32] = EVENT_TYPE_POINT;
                    instr_buf_wdata[31:0] = EVENT_INSTEVLOST;
                end else begin
                    instr_buf_wdata[39:0] = type_and_id;
                end
            end

            2'd1: begin
                instr_buf_wdata[COUNTER_WIDTH-1:0] = timestamp;
            end

            default: begin
                instr_buf_wdata = instr_event_data;
            end

        endcase
    end

    // Structure of in data:
    // If bit 104 is 0 -> the data represents an instrumentation setup where:
    //   - data[63,  0] is the instrumentation buffer offset
    //   - data[79, 64] is the number of slots pointer by the instrumentation offset
    // If bit 104 is 1 -> the data represents an instrumentation event where:
    //   - data[63,  0] is value field of the instrumentation event
    //   - data[103,64] is typeAndId field of the instrumentation event

    always_ff @(posedge clk) begin

        if (instr_buf_bvalid) begin
            bcount <= bcount - 2'd1;
        end

        case (state)

            IDLE: begin
                bcount <= 2'd3;
                write_buffer_count <= 2'd0;
                instr_event_data <= fifo_tdata[63:0];
                type_and_id <= fifo_tdata[103:64];
                timestamp <= fifo_timestamp;
                data_sent <= 1'b0;
                addr_sent <= 1'b0;
                if (fifo_tvalid) begin
                    if (fifo_tdata[104] == 1'b0) begin
                        overflow <= 1'b0;
                        slots <= fifo_tdata[64 +: EVENT_BUF_BITS+1];
                        current_slot <= {EVENT_BUF_BITS{1'b0}};
                        read_slot <= {EVENT_BUF_BITS{1'b0}};
                        avail_slots <= fifo_tdata[64 +: EVENT_BUF_BITS+1];
                        instr_buffer_addr <= fifo_tdata[AXI_ADDR_WIDTH-1:3]; //address must be aligned to 8 bytes
                    end else begin
                        if (WAIT_FOR_AVAIL_SLOT) begin
                            if (avail_slots == {EVENT_BUF_BITS+1{1'b0}}) begin
                                state <= READ_BUFFER_AR;
                            end else begin
                                state <= WRITE_BUFFER;
                            end
                        end else begin
                            if (avail_slots == EVENT_BUF_ONE || overflow) begin
                                state <= READ_BUFFER_AR;
                            end else begin
                                state <= WRITE_BUFFER;
                            end
                        end
                    end
                end
            end

            WRITE_BUFFER: begin
                if (instr_buf_awready && instr_buf_wready || (data_sent && instr_buf_awready) || (addr_sent && instr_buf_wready)) begin
                    write_buffer_count <= write_buffer_count + 2'd1;
                    data_sent <= 1'b0;
                    addr_sent <= 1'b0;
                    if (write_buffer_count == 2'd2) begin
                        state <= UPDATE_SLOTS;
                    end
                end else if (instr_buf_wready) begin
                    data_sent <= 1'b1;
                end else if (instr_buf_awready) begin
                    addr_sent <= 1'b1;
                end
            end

            UPDATE_SLOTS: begin
                current_slot <= current_slot + EVENT_BUF_ADDR_ONE == slots ? {EVENT_BUF_BITS{1'b0}} : current_slot + EVENT_BUF_ADDR_ONE;
                avail_slots <= avail_slots - EVENT_BUF_ONE;
                state <= WAIT_WRESP;
            end

            WAIT_WRESP: begin
                if (bcount == 2'd0) begin
                    state <= IDLE;
                end
            end

            READ_BUFFER_AR: begin
                if (instr_buf_arready) begin
                    state <= WAIT_READ_BUFFER;
                end
            end

            WAIT_READ_BUFFER: begin
                if (instr_buf_rvalid) begin
                    if (instr_buf_rdata[33:32] == EVENT_TYPE_INVALID) begin
                        avail_slots <= avail_slots + EVENT_BUF_ONE;
                        read_slot <= (read_slot + EVENT_BUF_ADDR_ONE) == slots ? {EVENT_BUF_BITS{1'b0}} : (read_slot + EVENT_BUF_ADDR_ONE);
                    end
                    if (instr_buf_rdata[33:32] != EVENT_TYPE_INVALID || (avail_slots + EVENT_BUF_ADDR_ONE == slots)) begin
                        state <= CHECK_AVAIL_SLOTS;
                    end else begin
                        state <= READ_BUFFER_AR;
                    end
                end
            end

            CHECK_AVAIL_SLOTS: begin
                if (WAIT_FOR_AVAIL_SLOT) begin
                    if (avail_slots != {EVENT_BUF_BITS+1{1'b0}}) begin
                        state <= WRITE_BUFFER;
                    end else begin
                        state <= READ_BUFFER_AR;
                    end
                end else begin
                    if (avail_slots > EVENT_BUF_ONE) begin
                        overflow <= 1'b0;
                        state <= WRITE_BUFFER;
                    //If we are in overflow, we need at least two slots,one for the current event and another one for a possible overflow
                    end else if (!overflow && avail_slots == EVENT_BUF_ONE) begin
                        overflow <= 1'b1;
                        state <= WRITE_BUFFER;
                    end else begin
                        state <= IDLE;
                    end
                end
            end

        endcase

        if (!rstn) begin
            state <= IDLE;
        end
    end

endmodule
